import copreus.drivers
import importlib
import pelops.mylogger as mylogger


class DeviceFactory:
    @staticmethod
    def create_devices(config, mqtt_client=None, logger=None, spi_lock=None):
        """Static devices factory - takes a list of device configs and creates them."""

        factory_logger = mylogger.get_child(logger, __name__)
        factory_logger.info("creating devices")
        factory_logger.debug("device configs: ".format(config["devices"]))
        devices = []
        for device_config in config["devices"]:
            if not device_config["active"]:
                continue
            config["device"] = device_config
            device = DeviceFactory.create_device(config, mqtt_client, logger, spi_lock)
            devices.append(device)
            factory_logger.info(" - added device '{}.{}'".format(device._type, device._name))
        return devices


    @staticmethod
    def create_device(config, mqtt_client=None, logger=None, spi_lock=None):
        """Static device factory - takes a device entry from json/yaml config and instantiates the corresponding
        Class. Classes that are specializations of ASPI are provided with the spi_lock (if one is provided to this
        factory).

        New implemented driver must be added manually."""
        type_name = config["device"]["type"].upper()

        # it is on purpose that not class names are used to be compared with type_name (as will be done within the
        # constructor of the base class ADriver). This approach allows for late binding - thus, a class is imported
        # if and only if it is needed which results in less dependencies that must be fullfilled although they might
        # not be needed.

        drivers = copreus.drivers.get_drivers()

        try:
            driver = drivers[type_name]
            mod = importlib.import_module(driver["module"])
            klass = getattr(mod, driver["name"])
            if "ASPI" in driver["bases"]:
                result = klass(config, mqtt_client, logger, spi_lock)
            else:
                result = klass(config, mqtt_client, logger)
        except:
            logger = mylogger.get_child(logger, __name__)
            logger.error("unknown type name '{}'.".format(type_name))
            raise ValueError("unknown type name '{}'.".format(type_name))

        return result

    @staticmethod
    def old_create(config, mqtt_client, logger, spi_lock=None):
        type_name = config["type"].upper()
        if type_name == "ADC":
            from copreus.drivers.adc import ADC
            result = ADC(config, mqtt_client, logger, spi_lock)
        elif type_name == "DAC":
            from copreus.drivers.dac import DAC
            result = DAC(config, mqtt_client, logger, spi_lock)
        elif type_name == "BME_280":
            from copreus.drivers.bme_280 import BME_280
            result = BME_280(config, mqtt_client, logger)
        elif type_name == "DHT":
            from copreus.drivers.dht import DHT
            result = DHT(config, mqtt_client, logger)
        elif type_name == "EPAPERDIRECT":
            from copreus.drivers.epaperdirect import EPaperDirect
            result = EPaperDirect(config, mqtt_client, logger, spi_lock)
        elif type_name == "EPAPERSIMPLE":
            from copreus.drivers.epapersimple import EPaperSimple
            result = EPaperSimple(config, mqtt_client, logger, spi_lock)
        elif type_name == "INPUT":
            from copreus.drivers.input import Input
            result = Input(config, mqtt_client, logger)
        elif type_name == "OUTPUT":
            from copreus.drivers.output import Output
            result = Output(config, mqtt_client, logger)
        elif type_name == "ROTARYENCODER":
            from copreus.drivers.rotaryencoder import RotaryEncoder
            result = RotaryEncoder(config, mqtt_client, logger)
        else:
            logger = mylogger.get_child(logger, __name__)
            logger.error("unknown type name '{}'.".format(type_name))
            raise ValueError("unknown type name '{}'.".format(type_name))

        return result





import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import _thread
from time import sleep, strftime
from copreus.drivers.dac import DAC
from tests_manual.tools import get_device_config, get_topics_sub, send_to_topic

filename = 'config_dac.yaml'

#send_to_topic(filename, "/test/closed", 0)

# check if use-calibration is set to 'false' ----------
c = get_device_config(filename)
if c["use-calibration"]:
    raise ValueError("for calibration parameter 'use-calibration' must be set to false "
                     "(currently: {}).".format(c["use-calibration"]))
# -----------------------------------------------------

# create csv file -------------------------------------
data_dac = open("calibration_dac_"+strftime("%Y%m%d-%H%M%S")+".csv", "w")
# -----------------------------------------------------


# start ADC driver ------------------------------------
_adc_thread = _thread.start_new_thread(DAC.standalone, (['-c', filename],))
topic_volt = get_topics_sub(filename)["volt"]
# -----------------------------------------------------


value = -1.0
counter = 0
max = c["maxvalue"]
min = 0
step = 0.5

# get user input for parameters ---------------------
sleep(1)  # wait a little bit - previous tasks should be finished before continuing
print("Default calibration steps: {} V to {} V with step width {} V.".format(min, max, step))
newmin = input("Min value ({} V): ".format(min))
try:
    if float(newmin) > min and float(newmin) < max:
        min = float(newmin)
        print("new value for min: {} V.".format(str(min)))
except ValueError:
    pass
try:
    newmax = input("Max value ({} V): ".format(max))
    if float(newmax) > min and float(newmax) <= max:
        max = float(newmax)
        print("new value for max: {} V.".format(max))
except ValueError:
    pass
try:
    newstep = input("Step value ({} V): ".format(step))
    if float(newstep) >= 0 and float(newstep) <= (max-min):
        step = float(newstep)
        print("new value for step: {} V.".format(step))
except ValueError:
    pass
# -----------------------------------------------------

print("Values used for calibration steps: {} V to {} V with step width {} V.".format(min, max, step))

print("start calibration -------------------------------")
print("volt_out;volt_in;multimeter")
print("# - [ref_value, raw_value]", file=data_dac)

for value in range(int(min*100), int(max*100)+1, int(step*100)):
    if value > int(max*100):
        break
    volt_out = value / 100.0
    print(" ... sending value twice to dac.")
    send_to_topic(filename, topic_volt, volt_out)
    sleep(1)
    send_to_topic(filename, topic_volt, volt_out)
    sleep(2)
    multimeter = -1
    while multimeter < 0:
        mm = input("multimeter value for out value {} V: ".format(volt_out))
        try:
            mm = float(mm)
            accept_value = input("accept entered value {} V (y/N)? ".format(mm))
            if accept_value == "y" or accept_value == "Y":
                multimeter = mm
        except ValueError:
            pass
    print("{};{}".format(volt_out, multimeter))
    # - [ref_value, raw_value]
    print(" - [{}, {}]".format(multimeter, volt_out), file=data_dac)

data_dac.close()
print("done. -------------------------------------------")
print("cleaning up.")
try:
    _thread.interrupt_main()
except KeyboardInterrupt:
    pass
sleep(1)
print("finished.")



import copreus.schema.adriver as adriver
import copreus.schema.apolling as apolling
import copreus.schema.aspi as aspi
import copreus.schema.calibratedvalue as calibratedvalue


def _add_schema_part(schema, schema_part):
    schema["device"]["required"].extend(schema_part["required"])
    schema["device"]["properties"].update(schema_part["properties"])


def get_schema():
    driver_specific_properties = {
        "maxvalue": {
            "description": "maximum value in volt. result will be normalized towards this value.",
            "type": "number"
        },
        "bit": {
            "description": "how many bits are used. typical values are 8, 10, and 12.",
            "type": "integer"
        },
    }

    topics_pub = {
        "raw": "raw (=integer value returned from adc)",
        "volt": "volt (=converted and calibrated value)"
    }

    apolling_schema_parts, topics_sub, mqtt_translations = apolling.get_schema_parts()
    schema = adriver.get_schema("adc", driver_specific_properties, topics_pub, topics_sub, mqtt_translations)
    _add_schema_part(schema, apolling_schema_parts)
    _add_schema_part(schema, aspi.get_schema_parts())
    _add_schema_part(schema, calibratedvalue.get_schema_parts("calibration"))

    return schema



import Adafruit_DHT
from copreus.baseclasses.adriver import ADriver
from copreus.baseclasses.apolling import APolling
from copreus.baseclasses.calibratedvalue import CalibratedValue
from copreus.schema.dht import get_schema


class DHT(ADriver, APolling):
    """Driver for the DHT temperature/humidity sensor family.

    The driver entry in the yaml file consists of:
      * ADriver entries
        * topics_pub: temperature, humidity
      * APolling entries
      * CalibratedValue entries in a sub-block named 'calibration_temperature'
      * CalibratedValue entries in a sub-block named 'calibration_humidity'
      * DHT entries
        * pin: gpio @ raspberry
        * sensor-type: DHT11, DHT22, AM2302

    Example:
    device:
        type: dht
        sensor-type: DHT22
        poll-interval: 30
        pin: 26
        topics-pub:
            temperature: /dht22/temperature/raw
            humidity: /dht22/humidity/raw
        topics-sub:
            poll-now: /dht22/pollnow
        mqtt-translations:
            poll-now: True
        use-calibration-temperature: True
        calibration-temperature:
        # - [ref_value, raw_value]
        use-calibration-humidity: True
        calibration-humidity:
        # - [ref_value, raw_value]
    """

    _pin = -1  # gpio pin id
    _sensor_type = None  # sensor type (one of the values in dict _sensor_type_list)
    _sensor_type_list = {  # list of valid sensor types. the sensor-type entry in yaml must be one of the keys.
        "DHT11": Adafruit_DHT.DHT11,
        "DHT22": Adafruit_DHT.DHT22,
        "AM2302": Adafruit_DHT.AM2302,
    }
    _calibrated_t = None  # copreus.baseclasses.CalibratedValue for temperature
    _calibrated_h = None  # copreus.baseclasses.CalibratedValue for humidity

    def __init__(self, config, mqtt_client=None, logger=None):
        ADriver.__init__(self, config, mqtt_client, logger, logger_name=__name__)
        APolling.__init__(self, self._config, self._mqtt_client, self._logger)
        self._pin = self._config["pin"]

        if self._config["sensor-type"] not in self._sensor_type_list.keys():
            self._logger.error("Wrong parameter. Value for 'sensor-type': {} is not in list of accepted values {}.".
                               format(self._config["sensor-type"], self._sensor_type_list.keys()))
            raise ValueError("Wrong parameter. Value for 'sensor-type': {} is not in list of accepted values {}.".
                             format(self._config["sensor-type"], self._sensor_type_list.keys()))
        self._sensor_type = self._sensor_type_list[self._config["sensor-type"]]

        self._calibrated_t = CalibratedValue(self._logger, self._config["calibration-temperature"], 1)
        self._calibrated_h = CalibratedValue(self._logger, self._config["calibration-humidity"], 1)

    def _poll_device(self):
        """APolling._poll_device"""
        humidity, temperature = None, None
        while humidity is None or temperature is None:
            humidity, temperature = Adafruit_DHT.read_retry(self._sensor_type, self._pin)

        t = self._calibrated_t.value(temperature)
        h = self._calibrated_t.value(humidity)

        self._publish_value(self._topics_pub["temperature"], t)
        self._publish_value(self._topics_pub["humidity"], h)

    def _driver_start(self):
        """ADriver._driver_start"""
        self._start_polling()

    def _driver_stop(self):
        """ADriver._driver_stop"""
        self._stop_polling()

    @classmethod
    def _get_schema(cls):
        return get_schema()


def standalone():
    """Calls the static method DHT.standalone()."""
    DHT.standalone()


if __name__ == "__main__":
    DHT.standalone()


from threading import Lock
from pelops.abstractmicroservice import AbstractMicroservice
from copreus.devicemanager.devicefactory import DeviceFactory
from copreus.schema.devicemanager import get_schema
from copreus import version


class DeviceManager(AbstractMicroservice):
    """Takes a yaml config file, creates all devices that are set to active, and starts them.

    The DeviceManager alters the behavior of the devices at three points:
      * A single instance of MyMQTTClient is provided to all driver.
      * It overrides the individual topic_sub_handler and provides one central _on_message handler.
      * One spi lock is provided to avoid overallocation of the spi interface (Please be note that a single lock is
      used for all spi instances independet of the bus/devices parameter. Thus, even in case of two independent spi
      interfaces only one of them can be used at any given time.)

    See copreus.baseclasses.adriver for a brief description of the yaml config file."""

    _version = version

    _devices = None  # list of instantiated devices.
    _spi_lock = None  # threading.Lock

    def __init__(self, config, mqtt_client=None, logger=None):
        AbstractMicroservice.__init__(self, config, "devices", mqtt_client, logger)
        self._spi_lock = Lock()
        # create_devices needs the full config structure - thus, config instead of self._config is used.
        self._devices = DeviceFactory.create_devices(config, self._mqtt_client, self._logger, self._spi_lock)

    def _start(self):
        """Starts all active devices."""
        print("DeviceManager started")
        for device in self._devices:
            device.start()

    def _stop(self):
        """Stops all active devices."""
        for device in self._devices:
            device.stop()
        print("DeviceManager stopped")

    @classmethod
    def _get_description(cls):
        return "Device Manager\n" \
               "In Greek mythology, Copreus (Κοπρεύς) was King Eurystheus' herald. He announced Heracles' Twelve " \
               "Labors. This script starts severaö device driver on a raspberry pi and connects them to MQTT. " \
               "Thus, copreus takes commands from the king (MQTT) and tells the hero (Device) what its labors " \
               "are. Further, copreus reports to the king whatever the hero has to tell him."

    @classmethod
    def _get_schema(cls):
        return get_schema()


def standalone():
    """Calls the static method DeviceManager.standalone()."""
    DeviceManager.standalone()


if __name__ == "__main__":
    DeviceManager.standalone()



from pelops import mylogger, mymqttclient, myconfigtools
import unittest
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))


class TestCalibratedValue(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._config = myconfigtools.read_config(
            os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))+"/tests_unit/config_calibratedvalue.yaml")
        cls.logger = mylogger.create_logger(cls._config["logger"], __name__)
        cls.logger.info("TestCalibratedValue - start")

    @classmethod
    def tearDownClass(cls):
        cls.logger.info("TestCalibratedValue - stop")



if __name__ == '__main__':
    unittest.main()



import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from pelops.mymqttclient import MyMQTTClient
from pelops.mylogger import create_logger

from yaml import load

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper
import os


def _on_message(msg):
    print("test - received message: '{}'".format(msg))


def get_logger(filename):
    config = load(open(filename, 'r'), Loader=Loader)
    logger = create_logger(config["logger"], __name__)
    return logger


def resister_to_pubs(filename, pos=None, on_message=None):
    mqtt = MyMQTTClient(get_mqtt_config(filename), get_logger(filename))
    if on_message is None:
        on_message = _on_message
    mqtt.connect()
    subs = get_topics_pub(filename, pos)
    for k,s in subs.items():
        print("registering topic '{}'.".format(s))
        mqtt.subscribe(s, on_message)


def send_to_topic(filename, topic, msg):
    print("sending to topic '{}' the value '{}'".format(topic, msg))
    mqtt = MyMQTTClient(get_mqtt_config(filename), get_logger(filename))
    mqtt.on_message = _on_message
    mqtt.connect()
    mqtt.publish(topic, msg)
    mqtt.disconnect()


def get_mqtt_config(filename):
    config = load(open(filename, 'r'), Loader=Loader)
    credentials = load(open(os.path.expanduser(config["mqtt"]["credentials-file"]), 'r'), Loader=Loader)
    config["mqtt"].update(credentials["mqtt"])
    return config["mqtt"]


def get_mqtt(filename):
    mqtt = MyMQTTClient(get_mqtt_config(filename), get_logger(filename))
    return mqtt


def get_topics_pub(filename, pos=None):
    c = get_device_config(filename, pos)
    pubs = c["topics-pub"]
    result = {}
    for k,v in pubs.items():
        result[k] = v
    return result


def get_resolution(filename, pos=None):
    c = get_device_config(filename, pos)
    return 2**c["bit"]


def get_maxvalue(filename, pos=None):
    c = get_device_config(filename, pos)
    return c["maxvalue"]


def get_device_config(filename, pos=None):
    config = load(open(filename, 'r'), Loader=Loader)
    if pos is None:
        c = config["device"]
    else:
        c = config["devices"][pos]
    return c


def get_topics_sub(filename, pos=None):
    c = get_device_config(filename, pos)
    subs = c["topics-sub"]
    result = {}
    for k,v in subs.items():
        result[k] = v
    return result


def get_mqtt_translations(filename, pos=None):
    c = get_device_config(filename, pos)
    subs = c["mqtt-translations"]
    return subs


def get_model(filename, pos=None):
    c = get_device_config(filename, pos)
    return str(c["model"])



import RPi.GPIO as GPIO


class AEvents(object):
    """Additional base class for driver that must react to changing pin states.

    Events are registered for BOTH (raising and falling edge). If necessary, the handler function must differentiate
    between raising and falling edge. Usually done by reading the current state of the pin.

    GPIO.setmode(GPIO.BCM)"""

    _events = None  # dict containing the pin-method parings.
    _logger_aevents = None  # print debugging information if set to yes.

    def __init__(self, config, logger):
        self._events = {}
        self._logger_aevents = logger
        GPIO.setmode(GPIO.BCM)

    def _add_event(self, pin, func, bounce_time=50):
        """Add an event handler """
        self._events[int(pin)] = {"f": func, "b": bounce_time}

    def _register_events(self):
        """Register all events that are stored in the dict _events. Usually called in the _start_sequence method of the
        silbling."""
        self._logger_aevents.info("AEvents._register_events - register events")
        for pin,event in self._events.items():
            self._logger_aevents.debug("AEvents._register_events - register event pin:{}, callback:'{}', "
                                      "bounce_time:'{}'".
                                      format(str(pin), str(event["f"]), str(event["b"])))
            GPIO.add_event_detect(pin, GPIO.BOTH, callback=event["f"], bouncetime=event["b"])

    def _unregister_events(self):
        """Unregister from all events that are stored in the dict _events. Usually called in the
         _stop_sequence method of the silbling."""
        self._logger_aevents.info("AEvents._unregister_events - unregister events")
        for pin,event in self._events.items():
            self._logger_aevents.debug("AEvents._unregister_events - remove event for pin {}.".format(pin))
            GPIO.remove_event_detect(pin)



import copreus.schema.adriver as adriver
import copreus.schema.apolling as apolling
import copreus.schema.aspi as aspi
import copreus.schema.calibratedvalue as calibratedvalue


def _add_schema_part(schema, schema_part):
    schema["device"]["required"].extend(schema_part["required"])
    schema["device"]["properties"].update(schema_part["properties"])


def get_schema():
    driver_specific_properties = {
        "pin": {
            "description": "gpio @ raspberry.",
            "type": "integer"
        },
        "sensor-type": {
            "description": "DHT11, DHT22, AM2302",
            "type": "string",
            "enum": ["DHT11", "DHT22", "AM2302"]
        },
    }

    topics_pub = {
        "temperature": "raw temperature",
        "humidity": "raw humidity"
    }

    apolling_schema_parts, topics_sub, mqtt_translations = apolling.get_schema_parts()
    schema = adriver.get_schema("dht", driver_specific_properties, topics_pub, topics_sub, mqtt_translations)
    _add_schema_part(schema, apolling_schema_parts)
    _add_schema_part(schema, calibratedvalue.get_schema_parts("calibration-humidity"))
    _add_schema_part(schema, calibratedvalue.get_schema_parts("calibration-temperature"))

    return schema



import os
import glob
import ast
import _ast
import copreus

version = copreus.version


def get_drivers(baseclasses=["ADriver", "AEPaper"]):
    """Get all classes and their modules that are silbilings of baseclasses.adriver.ADriver."""

    def extract_class_info(class_entry, filename, package, child_of):
        result = {
            "name": class_entry.name,
            "path": filename,
            "module": package + "." + filename[:-3].split('/')[-1],
            "bases": []
        }
        for b in class_entry.bases:
            result["bases"].append(b.id)

        found = False
        for co in child_of:
            if co in result["bases"]:
                found = True
                break
        if not found:
            result = None

        return result

    path = os.path.abspath(os.path.dirname(__file__))
    files = glob.glob(path + "/*.py")
    class_list = {}
    for filename in files:
        with open(filename, 'r') as myfile:
            sourcecode = myfile.read()
        tree = ast.parse(sourcecode)
        classes = [cls for cls in tree.body if isinstance(cls, _ast.ClassDef)]
        for c in classes:
            ci = extract_class_info(c, filename, 'copreus.drivers', child_of=baseclasses)
            if ci:
                class_list[ci["name"].upper()] = ci

    return class_list



import bme280 as bme280_driver  # https://pypi.python.org/pypi/RPi.bme280
from smbus2 import SMBus  # https://pypi.python.org/pypi/smbus2/0.2.0
from copreus.baseclasses.adriver import ADriver
from copreus.baseclasses.apolling import APolling
from copreus.baseclasses.calibratedvalue import CalibratedValue
from copreus.schema.bme_280 import get_schema


class BME_280(ADriver, APolling):  # BME280 would result in a name conflict with the above imported driver ...
    """Driver for the BME280 sensor (temperature, humidity, and pressure) with i2c connectivity.

    The driver entry in the yaml file consists of:
      * ADriver entries
        * topics_pub: temperature, humidity, pressure
      * APolling entries
      * i2c entries:
        * port: i2c port
        * address: i2c address
      * CalibratedValue entries in a sub-block named 'calibration_temperature'
      * CalibratedValue entries in a sub-block named 'calibration_humidity'
      * CalibratedValue entries in a sub-block named 'calibration_pressure'

    Example:
    device:
        poll-interval: 30
        type: bme_280
        port: 1
        address: 0x76
        topics-pub:
            temperature: /bme280/temperature/raw
            humidity: /bme280/humidity/raw
            pressure: /bme280/pressure/raw
        topics-sub:
            poll-now: /bme280/pollnow
        mqtt-translations:
            poll-now: True
        calibration-temperature:
            use-calibration: True
            # - [ref_value, raw_value]
        calibration-humidity:
            use-calibration: True
            # - [ref_value, raw_value]
        calibration-pressure:
            use-calibration: True
            # - [ref_value, raw_value]


    Note: this class is named BME_280 due to a name conflict with the used Adafruit driver named bme280.
    """

    _port = -1  # i2c port
    _address = -1  # i2c address
    _bus = None  # SMBus instance
    _calibrated_t = None  # copreus.baseclasses.CalibratedValue for temperature
    _calibrated_h = None  # copreus.baseclasses.CalibratedValue for humidity
    _calibrated_p = None  # copreus.baseclasses.CalibratedValue for airpressure

    def __init__(self, config, mqtt_client=None, logger=None):
        ADriver.__init__(self, config, mqtt_client, logger, logger_name=__name__)
        APolling.__init__(self, self._config, self._mqtt_client, self._logger)

        self._port = self._config["port"]
        self._address = self._config["address"]

        self._calibrated_t = CalibratedValue(self._logger, self._config["calibration-temperature"], 1)
        self._calibrated_h = CalibratedValue(self._logger, self._config["calibration-humidity"], 1)
        self._calibrated_p = CalibratedValue(self._logger, self._config["calibration-pressure"], 1)

    def _poll_device(self):
        """APolling._poll_device"""
        data = bme280_driver.sample(self._bus, self._address)

        t = self._calibrated_t.value(data.temperature)
        h = self._calibrated_t.value(data.humidity)
        p = self._calibrated_t.value(data.pressure)

        self._publish_value(self._topics_pub["temperature"], t)
        self._publish_value(self._topics_pub["humidity"], h)
        self._publish_value(self._topics_pub["pressure"], p)

    def _driver_start(self):
        """ADriver._driver_start"""
        self._bus = SMBus(self._port)
        bme280_driver.load_calibration_params(self._bus, self._address)
        self._start_polling()

    def _driver_stop(self):
        """ADriver._driver_stop"""
        self._stop_polling()
        self._bus.close()

    @classmethod
    def _get_schema(cls):
        return get_schema()


def standalone():
    """Calls the static method BME_280.standalone()."""
    BME_280.standalone()


if __name__ == "__main__":
    BME_280.standalone()



import copreus.schema.adriver as adriver
import copreus.schema.apolling as apolling
import copreus.schema.calibratedvalue as calibratedvalue


def _add_schema_part(schema, schema_part):
    schema["device"]["required"].extend(schema_part["required"])
    schema["device"]["properties"].update(schema_part["properties"])


def get_schema():
    driver_specific_properties = {
        "port": {
            "description": "i2c port",
            "type": "number"
        },
        "address": {
            "description": "i2c address",
            "type": "number"
        },
    }

    topics_pub = {
        "temperature": "raw temperature",
        "humidity": "raw humidity",
        "pressure": "raw pressure"
    }

    apolling_schema_parts, topics_sub, mqtt_translations = apolling.get_schema_parts()
    schema = adriver.get_schema("bme_280", driver_specific_properties, topics_pub, topics_sub, mqtt_translations)
    _add_schema_part(schema, apolling_schema_parts)
    _add_schema_part(schema, calibratedvalue.get_schema_parts("calibration-temperature"))
    _add_schema_part(schema, calibratedvalue.get_schema_parts("calibration-humidity"))
    _add_schema_part(schema, calibratedvalue.get_schema_parts("calibration-pressure"))

    return schema



import copreus.schema.adriver as adriver
import copreus.schema.apolling as apolling
import copreus.schema.aspi as aspi
import copreus.schema.calibratedvalue as calibratedvalue
import copreus.schema.aepaper as aepaper


def _add_schema_part(schema, schema_part):
    schema["device"]["required"].extend(schema_part["required"])
    schema["device"]["properties"].update(schema_part["properties"])


def get_schema():
    driver_specific_properties = aepaper.get_schema()

    topics_pub = {
        "message_queue_size": "publishes the number of messages that wait to be processes.",
    }

    topics_sub = {
        "full_image": "a single image covering the whole display to be placed in the current buffer.",
        "partial_image": "list of image covering only parts of the display plus their position to be placed "
                      "into the current buffer.",
        "switch_frame": "if command 'switch_buffer' is received via this topic, switch the frame buffer."
    }

    mqtt_translations = {
        "switch_frame": "command to switch between the two frames."
    }

    schema = adriver.get_schema("epaperdirect", driver_specific_properties, topics_pub, topics_sub, mqtt_translations)
    _add_schema_part(schema, aspi.get_schema_parts())

    return schema



import RPi.GPIO as GPIO
from copreus.baseclasses.adriver import ADriver
from copreus.schema.output import get_schema


class Output(ADriver):
    """Generic driver that sets the given output pin.

    The driver entry in the yaml file consists of:
      * ADriver entries
        * topics_sub: closed - mqtt-translations.closed-true and mqtt-translations.closed-false
      * Output entries
        * pin: gpio pin
        * physical-closed: high/low - mapping between logcial states (closed/open) and physical output
        parameters (low/high)
        * initially-closed: True/False - defines if the output is opened or closed after start of driver

    Example:
    device:
        type: output
        pin: 21
        initially-closed: True
        physical-closed: high
        topics-sub:
            closed: /test/closed
        mqtt-translations:
            closed-true: ON
            closed-false: OFF
    """

    _pin = -1  # gpio pin id
    _gpio_closed = -1  # value to write to gpio pin for closing output (0/1)
    _gpio_opened = -1  # value to write to gpio pin for opening output (0/1)
    _initially_closed = False  # should the output be opened or closed after start

    def __init__(self, config, mqtt_client=None, logger=None):
        ADriver.__init__(self, config, mqtt_client, logger, logger_name=__name__)

        self._mqtt_client.subscribe(self._topics_sub["closed"], self._message_closed)

        self._pin = int(self._config["pin"])
        self._initially_closed = bool(self._config["initially-closed"])
        if str(self._config["physical-closed"].lower()) == "low":
            self._gpio_closed = 0
        elif str(self._config["physical-closed"].lower()) == "high":
            self._gpio_closed = 1
        else:
            self._logger.error("'physical-closed' - expected 'low'/'high' but received '{}'.".
                               format(self._config["physical-closed"].lower()))
            raise ValueError("'physical-closed' - expected 'low'/'high' but received '{}'.".
                             format(self._config["physical-closed"].lower()))
        self._gpio_opened = (self._gpio_closed + 1) % 2

    def _message_closed(self, msg):
        """on_message handler for topic sub 'closed'."""
        self._logger.info("received message '{}' on topic '{}'.".format(msg, self._topics_sub["closed"]))
        if str(msg) == str(self._mqtt_translations["closed-true"]):
            self._logger.info("writing '{}' to pin '{}'.".format(self._gpio_closed, self._pin))
            GPIO.output(self._pin, self._gpio_closed)
        elif str(msg) == str(self._mqtt_translations["closed-false"]):
            self._logger.info("writing '{}' to pin '{}'.".format(self._gpio_opened, self._pin))
            GPIO.output(self._pin, self._gpio_opened)
        else:
            raise ValueError("Output.'{}'.payload expects {} or {}. received '{}' instead.".format(
                msg.topic, self._mqtt_translations["closed-true"], self._mqtt_translations["closed-false"],
                msg))

    def _driver_start(self):
        """ADriver._driver_start"""
        GPIO.setmode(GPIO.BCM)
        GPIO.setup(self._pin, GPIO.OUT)
        if self._initially_closed:
            GPIO.output(self._pin, self._gpio_closed)
        else:
            GPIO.output(self._pin, self._gpio_opened)

    def _driver_stop(self):
        """ADriver._driver_stop"""
        GPIO.cleanup(self._pin)

    @classmethod
    def _get_schema(cls):
        return get_schema()


def standalone():
    """Calls the static method Output.standalone()."""
    Output.standalone()


if __name__ == "__main__":
    Output.standalone()
